let {
  THREE,
  Edge,
  Node,
  OrbitControls,
  d3,
  ee,
  DEF_NODE_COLOR,
  DEF_NODE_HIGHLIGHT,
  DEF_ADJACENT_HIGHLIGHT,
  DEF_NODE_HTEXTURE,
  DEF_NODE_TEXTURE,
  DEF_NODE_NO_HTEX,
  DEF_NODE_NO_TEX,
  DEF_NODE_NO_HCOLOR,
  DEF_NODE_NO_COLOR,
  DEF_NODE_NO_HIGHLIGHT,
  DEF_NODE_SHAPE,
  DEF_NODE_HSHAPE,
  DEF_NODE_SIZE,
  DEF_HIGHLIGHT_SIZE,
  DEF_LAYOUT,
  DEF_LINES,
  DEF_MAP,
  DEF_TEXT_SCALE,
  DEF_TEXT_SIZE,
  DEF_TEXT_X_OFFSET,
  DEF_TEXT_Y_OFFSET,
  DEF_TEXT_BACK_COLOR,
  DEF_TEXT_OPACITY,
  TEXT_WIREFRAME
} = require('./imports').default;

var Utils = require('./utils/utils');

var Frame = function (graph, options) {
  // Needed to reference "this" in functions
  var self = this;

  // Options
  this.inGraph = graph;
  this.options = options || {
    "layout": DEF_LAYOUT
  };
  this.layout_options = this.options.graphLayout || {};

  this.lineObjects = [];
  this.lineIndices = [];
  this.MAX_LINES = 100000;
  this.textureLoader = new THREE.TextureLoader();
  this.textureLoader.crossOrigin = true;
  this.oldIntersect = -1;
  this.selection = [];
  this.fov = 60;
  this.near = 1;
  this.far = 5000;
  this.mouseStart = new THREE.Vector3(0, 0, 0);
  this.mouseEnd = new THREE.Vector3(0, 0, 0);
  this.dragging = null;
  this.showBox = false;
  this.renderWidth = 500;
  this.renderHeight = 500;
  this.maxZoom = 1000;
  this.paused = false;
  this.mouseDown = false;
  this.ee = ee({});
  this.layoutInit = true;
  this.labelSize = 6;
  this.relativeFontSize = 1;
  this.mapShowing = DEF_MAP;
  this.darkMode = false;

  require('./setup')(this);
  require('./addremove')(this);
  require('./select')(this);
  require('./modify')(this);
  require('./draw')(this);
  require('./update')(this);

  /**
   *  Starting point, run once to create scene
   */
  this.display = function () {
    if (graph == null) {
      Utils.randomGraph(self);
    }
    this.animate();
  }

  /**
   *  Creates loop called on every animation frame
   */
  let fps = 48;
  let now;
  let then = Date.now();
  let interval = 1000/fps;
  let delta;
  this.animate = function () {
    requestAnimationFrame( self.animate );

    now = Date.now();
    delta = now - then;
    if (delta > interval) {
      then = now - (delta % interval);
      self.controls.update();
      self.render();
    }
  };

  /**
   *  Set initial properties
   */
  this.init = function () {
    self.renderer = new THREE.WebGLRenderer({
      alpha: true
    });
    self.setDisplayParams();
    self.setRendererParams();

    self.setupCamera();
    self.setupGeometry();
    self.setupSelect();

    self.element.appendChild(self.renderer.domElement);
    self.element.appendChild(self.cssRenderer.domElement);

    self.canvas = document.querySelector('graph-container');

    self.setupLayout();

    self.setupGraph();

    window.addEventListener('resize', function(e) {
      e.preventDefault();
      self.onWindowResize();
    }, false);
  };

  /**
   * Change camera on window resize
   */
  self.onWindowResize = function () {
    self.clientRect = self.element.getBoundingClientRect();
    self.width = self.clientRect.width;
    self.height = self.clientRect.height;
    self.aspect = self.width / self.height;
    self.ccamera.aspect = self.aspect;
    self.ccamera.updateProjectionMatrix();

    self.renderer.setSize(self.width, self.height);
    self.cssRenderer.setSize(self.width, self.height);
  }

  /**
   *  Draws graphics
   */
  var stage = 0;
  this.render = function () {
    self.updateCamera();
    self.updateNodes();
    if (stage == 1) {
      self.updateLabels();
      self.updateEdges();
      stage = 0;
    }
    stage += 1;
    if (self.options.layout == "d3") {
      if (self.layoutInit == true) {
        var nodes = []
        self.graph.forEachNode(function (node) {
          nodes.push(node);
        });
        self.force.nodes(nodes);
        self.force.force("link", d3.forceLink(self.lineIndices));
        self.force.restart();
        self.force.stop();
        if (self.tickToStatic) {
          // We don't tick and pause for now
          // for (var i = 0, n = Math.ceil(Math.log(self.force.alphaMin()) / Math.log(1 - self.force.alphaDecay())); i < n; ++i) {
          //   self.force.tick();
          // }
          self.pauseLayout();
        }
        self.layoutInit = false;
      }
    }
    self.renderer.setViewport(0, 0, 1 * self.width, 1 * self.height);
    self.renderer.setScissor(0, 0, 1 * self.width, 1 * self.height);
    self.renderer.setScissorTest(true);
    self.renderer.render(self.scene, self.ccamera);
    self.cssRenderer.render(self.scene, self.ccamera);
    if (self.mapShowing) {
      self.renderer.setViewport(0, 0, 0.2 * (self.height), 0.2 * (self.height));
      self.renderer.setScissor(0, 0, 0.2 * (self.height), 0.2 * (self.height));
      self.renderer.setScissorTest(true);
      self.renderer.render(self.scene, self.ccamera2);
    }
  };

  this.updateGraph = function (graph) {
    var numNodesAdded = 0;
    graph.forEachNode(function (node) {
      var oldNode = self.graph.getNode(node.id);
      if (!oldNode) {
        self.addNode(node);
        numNodesAdded += 1;
      } else {
        self.updateNode(oldNode, node);
      }
    });
    graph.forEachLink(function (link) {
      if (!self.graph.getLink(link.fromId, link.toId)) {
        self.addEdge(self.graph.getNode(link.fromId), self.graph.getNode(link.toId));
      }
    });

    self.graph.forEachNode(function (node) {
      var oldNode = graph.getNode(node.id);
      if (!oldNode) {
        self.removeNode(node);
      }
    });

    self.force.alpha(1);
    self.force.stop();

    if (numNodesAdded > 10) {
      self.tickToStatic = true;
    } else {
      self.tickToStatic = false;
    }
  }

  this.updatePositions = function (positions) {
    self.graph.forEachNode(function(node) {
      var pos = positions[node.id];
      if (pos) {
        self.graph.getNode(node.id).x = pos[0];
        self.graph.getNode(node.id).y = pos[1];
        self.graph.getNode(node.id).pinned = node.pinned;
        if (node.pinned) {
          self.graph.getNode(node.id).fx = pos[0];
          self.graph.getNode(node.id).fy = pos[1];
        }
      }
    });
  }

  this.onSelectNodes = function () {

  };

  this.pauseLayout = () => {
    self.paused = true;

  };
   window.pause = this.pauseLayout;

  this.resumeLayout = () => {
    self.paused = false;
  };

  this.mapSelectedNodes = () => {
    self.mapNodeAttributes(["degree", "log", "scalexy"], self.selection);
  };

  this.mapAllNodes = () => {
    self.mapNodeAttributes(["degree", "log", "scalexy"]);
  };

  this.pinSelectedNodes = () => {
    self.mapNodeAttributes([true, "", "pinned"], self.selection);
  };

  this.unpinSelectedNodes = () => {
    self.mapNodeAttributes([false, "", "pinned"], self.selection);
  };

  // Size in em, not px
  this.setLabelFontSize = (size) => {
    self.graph.forEachNode(function (node) {
      self.changeLabelFontSize(node, size, self.relativeFontSize);
    });
  }

  this.setLabelRelativeSize = (size) => {
    self.relativeFontSize = size;
    self.setLabelFontSize(self.labelSize);
  }

  this.setLabelLength = (numChars) => {
    self.graph.forEachNode(function (node) {
      self.changeLabelLength(node, numChars);
    })
  }

  this.toggleMap = () => {
    self.mapShowing = !self.mapShowing;
  }

  this.showSelectedLabels = () => {
    for (var i = 0; i < self.selection.length; i++) {
      self.selection[i].renderData.textHolder.children[0].element.override = true;
    }
  }

  this.hideSelectedLabels = () => {
    for (var i = 0; i < self.selection.length; i++) {
      self.selection[i].renderData.textHolder.children[0].element.override = false;
    }
  }

  this.hideAllLabels = () => {
    self.graph.forEachNode(function (node) {
      var node = self.graph.getNode(node.id);
      node.renderData.textHolder.children[0].element.override = false;
    });
  }

  this.updateNode = function (node, newNode) {
    if (newNode.data) {
      if (node.data.color && newNode.data.color) {
        node.data.color = newNode.data.color;
        node.renderData.color = node.data.color;
        node.renderData.draw_object.material.color.set(new THREE.Color(node.data.color));
      }
      if (newNode.data.size && newNode.data.size > 0) {
        node.data.size = newNode.data.size;
        node.renderData.size = newNode.data.size;
        node.renderData.draw_object.scale.set(newNode.data.size, newNode.data.size, 1);
        var hsize = 1 + (DEF_HIGHLIGHT_SIZE/newNode.data.size);
        node.renderData.draw_object.children[0].scale.set(hsize, hsize, 1);
      }
      if (newNode.data.label) {
        node.data.label = newNode.data.label;
        node.renderData.label = node.data.label;
        node.renderData.textHolder.children[0].element.childNodes[0].innerText = node.data.label;
      }
      if (newNode.data.shape) {
        node.data.shape = newNode.data.shape;
        node.renderData.shape = node.data.shape;
        self.setNodeShape(node, node.data.shape);
      }
    }
  }

  this.setCanvasSize = function(size) {
    self.setBoundarySize(size);
  }

  this.setAllNodesShape = function(shape) {
    if (self.selection.length == 0) {
      self.graph.forEachNode(function (node) {
        self.setNodeShape(self.graph.getNode(node.id), shape);
      });
    } else {
      self.selection.forEach(function(node) {
        self.setNodeShape(self.graph.getNode(node.id), shape);
      });
    }
  }

  this.setNodeShape = function(node, shape) {
    if (shape == "square") {
      node.renderData.draw_object.geometry = Utils.make2x2Rect();
      node.renderData.draw_object.children[0].geometry = Utils.make2x2Rect();
    } else if (shape == 'circle') {
      node.renderData.draw_object.geometry = new THREE.CircleGeometry(1, 32);
      node.renderData.draw_object.children[0].geometry = new THREE.CircleGeometry(1, 32);
    } else if (shape == 'triangle') {
      node.renderData.draw_object.geometry = new THREE.CircleGeometry(1, 3);
      node.renderData.draw_object.children[0].geometry = new THREE.CircleGeometry(1, 3);
    } else if (shape == 'pentagon') {
      node.renderData.draw_object.geometry = new THREE.CircleGeometry(1, 5);
      node.renderData.draw_object.children[0].geometry = new THREE.CircleGeometry(1, 5);
    } else if (shape == 'hexagon') {
      node.renderData.draw_object.geometry = new THREE.CircleGeometry(1, 6);
      node.renderData.draw_object.children[0].geometry = new THREE.CircleGeometry(1, 6);
    } else if (shape == 'octagon') {
      node.renderData.draw_object.geometry = new THREE.CircleGeometry(1, 8);
      node.renderData.draw_object.children[0].geometry = new THREE.CircleGeometry(1, 8);
    }
  }
  
  this.getNumSelected = function() {
    if (self.selection.length == self.graph.getNodesCount()) {
      return 0;
    }
    return self.selection.length;
  }

  this.toggleDark = function() {
    self.darkMode = !self.darkMode;
  }

  this.getGraph = function() {
    return self.graph
  }

  this.getNodeList = function() {
    var nodes = [];
    self.graph.forEachNode(function(node) {
      nodes.push(node);
    });
    return nodes
  }

  this.getPositions = function() {
    var poses = {};
    self.graph.forEachNode(function(node) {
      poses[node.id] = [node.x, node.y];
    });
    return poses
  }

  this.getPinPositions = function() {
    var poses = {};
    self.graph.forEachNode(function(node) {
      poses[node.id] = [node.fx, node.fy];
    });
    return poses
  }
};

exports.Frame = Frame;
